! Information on the  job
! This includes  what job is being run
! Thread number OMP/MPI,etc
module wei_run
  use wei_external
  use wei_util
#ifdef SEP_MPI
  use mpi_sep 
#endif
#ifdef SEP_TIMER
  use sep_timers_mod
#endif
use sep
  implicit none
  integer,private,save :: mem_usage
  complex , pointer  :: small_buf(:,:)
  logical,save :: wei_dfile,wei_rfile



  !we will think of these all as public variable
  type weirun
     character(len=4)   :: tag_num
     character(len=9)   :: mach
     character(len=128) :: operator,logfile
     character(len=128) :: operation,file_out,local_path
!     character(len=128) :: operation,file_out,local_path,final_res_path
     logical            :: verb_all !! verbose
     logical            :: verb    !! verbose
     logical            :: verb_mem_usage !!
     logical            :: debug   !! Debug
     logical            :: depar   !! debug in the parallel regions
     logical            :: restart !! whether or not we are restarting the job
     logical            :: omp,mpi !! whether or not this is an omp/mpi job 
     logical            :: share_image !! share image file
     integer            :: io_unit !! unit to write out verb info 
     integer            :: m_ith!! mpi thread number
     integer            :: m_nth !! number of mpi threads
     integer            :: o_nth !! number of omp threads
     integer            :: fk_extra !! number of extra wavefields copies needed by fk op
     integer            :: mem_save !! whether to try to save memory
                          !0-none(default),1-no slowness arrays
                          !2-1 extra wavefield 3-2-extra wavefields
     logical            :: unlink
     integer            :: init_t
     integer            :: io_t
     integer            :: distrib_t
     integer            :: kxmig_t
     integer            :: fk_t
     integer            :: fx_t
     integer            :: fft_t
     integer            :: fftw_t
     integer            :: image_t
     integer            :: slow_t
     integer            :: collect_t
     integer            :: s_dis_t
     integer            :: diskio_t
     integer            :: junk1_t,junk2_t,junk3_t,junk4_t,junk5_t,junk6_t
  end type weirun

   type(weirun) :: run

  contains
!----------------------------------------------------------------
!
!
! Function for initializing the run enviro
!
!

  logical function weirun_init()
    character(len=128) :: operator,operation
      character(len=1024) :: temp_ch,temp2_ch
     logical :: new_datapath
    integer :: nsize,ierr,id,ir
     real ::a

#ifdef DEBUG
    character(len=128), parameter :: unit='weirun_init'
    if(run%debug) call in(unit)
#endif DEBUG
  
    mem_usage=0

    !Initialize parallel environment for run
    if(.not. run_parallel()) then
       write(0,*) "trouble initializing the wei_run_parallel"
       return
    end if

    call from_param('operation',operation,'migration...')
    run%operation = operation
    run%logfile="logfile"
    if(run%operation(1:3)=="mig" .or. run%operation(1:3)=="inv") run%logfile="R"
    if(run%operation(1:3)=="mod") run%logfile="D"
    call from_param("logfile",run%logfile,run%logfile)


   id=0
   ir=0;
#ifdef SEP_MPI
if(run%m_ith==0) then
#endif
     if(1==getch("output_datapath","s",temp2_ch)) then
       new_datapath=.true.
       temp_ch="datapath="//trim(temp2_ch)
       call getch_add_string(temp_ch)
     else
       new_datapath=.false.
     end if

     if(isfile("D"))  then
       call auxinout("D")
       id=1
     end if
     if(isfile("R"))  then
       call auxinout("R")
       ir=1
     end if
     call set_yes_putch()
     call set_alternate_putch(run%logfile)
     if((trim(run%logfile)=="D" .and. id==0) .or. &
       (trim(run%logfile)=="R" .and. ir==0) ) then
       call auxout(trim(run%logfile))
       if(new_datapath) ierr=srite(trim(run%logfile),a,4)
       call auxclose(trim(run%logfile))
       call auxinout(trim(run%logfile))
     end if
     call to_history("operation",run%operation,run%logfile)
#ifdef SEP_MPI
end if
  call MPI_BCAST(ir,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(id,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
#endif
  wei_rfile=(ir==1)
  wei_dfile=(id==1)
  if(run%m_ith /=0 ) call auxinout("R")
  if(run%m_ith /=0 ) call auxinout("D")


#ifdef SEP_MPI
  write(temp_ch,"(a,a,a,a)")"datapath=",trim(run%local_path),trim(run%tag_num),&
       "DATA_"
  call getch_add_string(temp_ch)
#endif


    call from_param('operator',operator,'.....')
    run%operator = operator

    call from_param('mem_monitor', run%verb_mem_usage, F)
    call from_param('verb', run%verb_all, F)
    call from_param('verb', run%verb, F)
    call from_param('debug',run%debug,F)
    call from_param('restart',run%restart,F)
    call from_param('depar',run%depar,F)
    call from_param("unlink",run%unlink,T)
    if(      run%debug) run%verb =T
    if(.not. run%debug) run%depar=T

    !! whether or not to save memory
    call from_param("mem_save",run%mem_save,0)
write(0,*)'mem_save=',run%mem_save
    call from_param("copy_buf",nsize,5);nsize=nsize*1000*1000/8
    allocate(small_buf(nsize,2))
     



    weirun_init=F

    call from_param('operator',operator,'.....')
    run%operator = operator

    if(run%mem_save > 0 .and. run%operation(1:3)/="mig" ) then
      write(0,*) "memory save option only available for migration"
      return
    end if 
      

   call from_param("io_file",run%file_out,"NONE")
   if(run%file_out(1:4)=="NONE") then
     run%io_unit=0
   else 
     run%io_unit=6
     open(run%io_unit,file=run%file_out)
   end if



   weirun_init=T

   if(run%m_ith.ne. 0) run%debug=.false.
   if(run%m_ith.ne. 0) run%verb=.false.
   if(run%verb) call run_report()

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function weirun_init


    

!----------------------------------------------------------------
!
!
! Subroutine to initialize OMP run environment
!
! Sets number of omp threads
!
!
!
!

  logical function run_omp_parallel()
#ifdef DEBUG
    character(len=128), parameter :: unit='omp_run_parallel'
    if(run%debug) call in(unit)
#endif

#ifdef SEP_OMP
    !$OMP PARALLEL
    run%o_nth = omp_get_num_threads()
    !$OMP END PARALLEL

    run%omp=T
#else
    run%o_nth=1
    run%omp=F
#endif

   run_omp_parallel=.true.



#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function  run_omp_parallel

!----------------------------------------------------------------
!
! Function for setting up mpi parallel enviro for mpi
!
! Turns of verb, debug except for the master thread
! Sets up the local datapath
! Figues out the extra to add for each thread
! records the number of threads and this threads num
!
 logical function runmpi_parallel()
   character(len=128) :: temp_ch,machine
   integer :: ierr,ilen,igrab,i
#ifdef DEBUG
    character(len=128), parameter :: unit='runmpi_parallel'
    if(run%debug) call in(unit)
#endif

    runmpi_parallel=F

#ifdef SEP_MPI
    call MPI_Get_processor_name(machine,ilen,ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, run%m_nth,ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, run%m_ith,ierr)
    call from_param("share_image",run%share_image,.false.)
#else
    run%share_image=.false.
    run%m_nth=1
    run%m_ith=0
#endif

    !TURN OFF VERB AND DEBUG FOR ALL BUT ZERO MPI THREAD
    if(run%m_ith /= 0) then
       run%verb=F
       run%debug=F
    end if


#ifdef SEP_MPI
    if(run%m_nth>1) then
!      do i=1,len(run%local_path)-1
      do i=1,len(run%local_path)
        run%local_path(i:i)=" "
      end do
!call MPI_Barrier(MPI_COMM_WORLD,ierr)
!     write(0,*) "#0 local_path",trim(run%local_path)
       igrab=getch("local_datapath","s",run%local_path)

!     write(0,*) "#1 local_path",trim(run%local_path)

!     write(0,*) "igrab",igrab

       if(0==igrab) run%local_path="/net/"//machine(1:ilen)//"/scr1/biondo/"

!     write(0,*) "#2 local_path",trim(run%local_path)

       run%mach=trim(machine(1:8))
    ilen=len(trim(run%local_path));
    if(igrab==1) run%local_path(ilen:ilen)=" "
       if(run%m_ith<10) then
          write(run%tag_num,"(a,i1)") "_",run%m_ith
       else if(run%m_ith <100) then
          write(run%tag_num,"(a,i2)") "_",run%m_ith
       else
          write(run%tag_num,"(a,i3)") "_",run%m_ith
       end if
    end if
    run%mpi=T
    if(run%m_nth==1) run%mpi=F
!     write(0,*) "#3 local_path",trim(run%local_path)
#else
    run%mpi=F
#endif


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif

  runmpi_parallel=T

  end function runmpi_parallel


 subroutine run_report()
    write(run%io_unit,*) "operator=",trim(run%operator(1:4))," operation=",trim(run%operation(1:8))
    write(run%io_unit,*) "writing to device ",run%io_unit
    write(run%io_unit,*) "verbose=",run%verb," debug=",run%debug," para debug", run%depar
    write(run%io_unit,*) "restart=",run%restart
    write(run%io_unit,*)  "mpi",run%mpi," mpi_ithread=",run%m_ith," mpi_nthread=",run%m_nth
    write(run%io_unit,*)  "omp",run%omp," omp_nthread=",run%o_nth
 end subroutine


!----------------------------------------------------------------
!
!  Subroutine for initializing parralel run (omp and mpi)
!
!
!
!
  logical function run_parallel()
     logical :: dude
#ifdef DEBUG
    character(len=128), parameter :: unit='run_parallel'
    if(run%debug) call in(unit)
#endif
 
    run_parallel=F

    if(.not. run_omp_parallel()) then
      write(0,*) "trouble initializing omp enviro"
      return
    end if

     dude=runmpi_parallel()
    if(.not. dude) then
      write(0,*) "trouble initializing mpi enviro"
      return
    end if
    
    run_parallel=T
#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function run_parallel
   
   subroutine del_mem(iadd)
    integer  :: iadd
      mem_usage=mem_usage-iadd        
   end subroutine
   subroutine show_mem(str)
    character(len=*) :: str
     if(run%verb .and. run%verb_mem_usage) write(0,*) "memory used (mb) ",trim(str),":",real(mem_usage)*8./1000000.
   end subroutine
   subroutine add_mem(iadd)
    integer  :: iadd
      mem_usage=mem_usage+iadd        
   end subroutine


end module wei_run
