module wei_var
  use wei_external
  use wei_down_type
  use wei_run
  use wei_lagrange
  use wei_var_type
  implicit none
  integer,private,save :: npart,part(1000),iz(1000)
   integer,  private, pointer, save :: igood_len(:),ihxbeg(:,:),ihybeg(:,:),ihxend(:,:),ihyend(:,:), do_freq_block(:)
  real, private, save,pointer :: zdepth(:,:),max_slow(:)

 contains

!-----------------------------------------------------
!
!
!
!
!
!


  logical function wei_var_init()
#ifdef DEBUG
    character(len=128), parameter :: unit='run_wei_var_init'
    if(run%debug) call in(unit)
#endif DEBUG


     allocate(var%dzb(down%nzs))
     call add_mem(size(var%dzb))
   wei_var_init=T

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif DEBUG

  end function



!-----------------------------------------------------
!
!
!   Initialize distribution pattern
!   (figure out what frequencies we are doing, where
!     they go, and what sampling we are using)
!
   logical function init_distrib_pattern()
     integer, pointer :: ifreq_list(:)
     real, pointer :: freq_list(:,:),wt_list(:,:)
     init_distrib_pattern=F
     
      if(.not. create_frequency_list(ifreq_list,freq_list,wt_list)) then
        write(0,*) "trouble creating frequency list"
        return 
      end if

      allocate(down%iw_thread(size(ifreq_list)))
     call add_mem(size(down%iw_thread))
     
      if(down%order(1:3)/="VAR") then
        if(.not. distribute_reg(ifreq_list,wt_list,freq_list)) then
           write(0,*) "trouble distributing regular"
           return 
         end if
      else 
        if(.not. distribute_var_freq_slow(ifreq_list,freq_list,wt_list)) then
           write(0,*) "trouble initializing varialbe freq slow pattern"
           return
        end if
      if(.not. init_reduce_h()) then
         write(0,*) "trouble in reduced h calculation"
         return
       end if
      end if

     if(.not. store_distrib_pattern()) then
       write(0,*) "trouble storing distribution pattern"
       return
     end if
     call del_mem(size(freq_list))
     call del_mem(size(ifreq_list))
     call del_mem(size(wt_list))
     deallocate(ifreq_list,freq_list,wt_list)

     init_distrib_pattern=T
     if(run%verb) write(0,*) "finished init distrib pattern "
   end function 

!-----------------------------------------------------
!
!
!  Create the list of frequencies we need to downward continue
!
!
logical function create_frequency_list(ifreq,freq_list,wt_list)
integer :: i,nredo,redo(10000),i_f,ib,n
integer, pointer :: ifreq(:)
real,pointer :: freq_list(:,:),wt_list(:,:),frac(:,:)
integer,allocatable :: index(:,:)
real, allocatable ::flist(:),wlist(:)
real :: f,o,d

create_frequency_list=F

  allocate(flist(down%dsep%n(5)))
  allocate(wlist(down%dsep%n(5)))
  call add_mem(size(flist)+size(wlist))



  call from_param("ireg_freq",var%ireg_freq,.false.)
  if(var%ireg_freq) then
     if(4*size(flist)/=sreed("freq_list",flist,size(flist)*4)) then
       write(0,*) "ireg_freq specified but freq_list did not contain valid freq_list"
       return
     end if
    call from_param("nw_sim",n)
    call from_param("ow_sim",o)
    call from_param("dw_sim",d)
    allocate(index(2,n),frac(2,n))
  call add_mem(size(index)+size(frac))
    call lagrange_init(index,frac,flist,n,o,d) 
    wlist=0. 
    do i_f=1,n
      do ib=1,size(index,1);
        wlist(index(ib,i_f))=wlist(index(ib,i_f))+frac(ib,i_f)
      end do
    end do
  call del_mem(size(index)+size(frac))
    deallocate(index,frac)
  else
    do i_f=1,down%dsep%n(5)
      flist(i_f)=down%dsep%o(5)+(i_f-1)*down%dsep%d(5)
    end do
    wlist=1.
  end if
    wlist=1.



if(.not. run%restart)  then 
  if(mod(down%dsep%n(5),down%nws)/=0) then 
    write(0,*) down%dsep%n(5),"=down%dsep%n(5) down%nws=",down%nws
    write(0,*) "nws must be a multiple of down%dsep%n(5)",mod(down%dsep%n(5),down%nws)
    return
  end if
 allocate(ifreq(down%dsep%n(5)/down%nws))
 allocate(wt_list(down%nws,size(ifreq)))
 allocate(freq_list(down%nws,size(ifreq)))
  call add_mem(size(ifreq)+size(wt_list)+size(freq_list))
 ib=0
 do i=1,size(ifreq)
   ifreq(i)=i
   do i_f=1,down%nws
     ib=ib+1
     freq_list(i_f,i)=flist(ib)
     wt_list(i_f,i)=wlist(ib)
   end do
 end do
else 
! nredo=getch("redo_groups","d",redo)
! i=getch("redo_part","d",iz)
! npart=getch("part_z","d",part)
! if(i /= npart) &
!   write(0,*) "the number of redo_part and part_z not equal"
!   
! allocate(ifreq(npart+nredo))
! ifreq(1:nredo)=redo(1:nredo)
! ifreq(nredo+1:nredo+npart)=part(1:npart)
  call seperr("not fixed")
end if

  
  call del_mem(size(flist)+size(wlist))
 deallocate(flist,wlist)
 down%nwp=size(ifreq)

create_frequency_list=T
end function
!-----------------------------------------------------
!
!
!  For the normal approach just distribute cycles
!
!
logical function distribute_reg(ifreq,wt_list,freq_list)
  integer, pointer :: ifreq(:)
  real :: wt_list(:,:),freq_list(:,:)
  integer :: iw,isend,i,ia,ib

  distribute_reg=F


  if(mod(size(ifreq),run%m_nth)/=0) then
    if(run%verb) write(0,*) "WARNING--m_nth not a multiple of size(ifreq)"
  end if

  down%nw_local=0;
  isend=0
  do iw=1,size(ifreq)
     down%iw_thread(iw)=isend
     if(run%debug) write(0,"(a,i3,a,i3)") "SENDING  thread block",iw,&
     " to  processor ", down%iw_thread(iw)
     if(isend==run%m_ith) down%nw_local=down%nw_local+1
     isend=isend+1
     if(isend==run%m_nth) isend=0
  end do

allocate(var%ihyb_beg(down%rsep%n(5)))
allocate(var%ihxb_beg(down%rsep%n(5)))
allocate(var%ihyb_end(down%rsep%n(5)))
allocate(var%ihxb_end(down%rsep%n(5)))
call add_mem(size(var%ihxb_beg)*4)


  !Record the first frequency in each group
  allocate(down%iw_first(down%nw_local))
  allocate(down%wts(down%nws,down%nw_local))
  allocate(down%freqs(down%nws,down%nw_local))
  allocate(down%w_local(down%nws))
  allocate(down%wt_local(down%nws))
   call add_mem(down%nw_local+down%nws*down%nw_local*2+down%nws*2)
  isend=0
  ia=0
  ib=0
  do iw=1,size(ifreq)
     ib=ib+1
     if(down%iw_thread(iw)==run%m_ith) then
        isend=isend+1
        down%iw_first(isend)=(ifreq(iw)-1)*down%nws+1
        do i=1,down%nws
           ia=ia+1
!           down%freqs(i,isend)=(down%iw_first(isend)-2+i)*down%dsep%d(5)+down%dsep%o(5)
           down%freqs(i,isend)=freq_list(i,ib)
           down%wts(i,isend)=wt_list(i,ib)
        end do
     end if
  end do
  distribute_reg=T
end function

!--------------------------------------------------------------
!
!  Distribute the frequencies with variable steps
!
!
!

logical function distribute_var_freq_slow(ifreq,freq_list,wt_list)
integer :: iw,i,k,iww
integer :: isend,nc,ierr,m,npts
integer :: ifreq(:)
real :: freq_list(:,:),wt_list(:,:)
real :: min_samp,max_samp,base_freq,io,lowcost
integer :: cycles,nsamp,samples,itry,inode(1),igroup,j,nw,hxk
real    ::  freq,lowest,myrand,mycost,min_zsamp,max_zsamp
real,allocatable    :: mywork(:),dz(:,:),cost_node(:),flist(:)
integer, allocatable   :: nz(:)
integer, allocatable   :: low_plan(:),cur_plan(:)
integer :: nz_mem
real :: mid
logical,allocatable :: not_distrib(:)
logical :: rite_sample
distribute_var_freq_slow=F

call from_param("min_zsamp",min_zsamp,down%rsep%d(5)/3.)
call from_param("max_zsamp",max_zsamp,down%rsep%d(5)*3)
call from_param("cycles",cycles,1000000)
call from_param("samp_per",samples,8)
call from_param("io_cost",io,20.)
isend=0
down%nw_local=0

nw=size(ifreq)
allocate(dz(100000,nw))
allocate(nz(nw))
allocate(not_distrib(nw))
allocate(low_plan(nw))
allocate(cur_plan(nw))
allocate(mywork(nw))
allocate(flist(nw))
allocate(cost_node(run%m_nth))
call add_mem(size(dz)+nw*6+size(cost_node))



!call return_max_slow(max_slow)
 max_slow=>down%max_slow

call get_mid_freq_list(ifreq,freq_list,flist)

do iw=1,nw
  freq=flist(iw)
  nz(iw)=calc_work(freq/6.285714,min_zsamp,max_zsamp,cycles,samples,dz(:,iw))
  mywork(iw)=io+nz(iw)
!  if(run%m_ith==0) write(0,*) "MY WORK",iw,nz(iw)
end do
call del_mem(size(flist))
deallocate(flist)

lowest=10000.

#ifdef SEP_MPI
    if(run%m_ith==0) then
#endif


!Find the plan that equalizes the work as much as possible
do itry=1,10*nw
  cost_node=0;

  not_distrib=.true.

  do while(any(not_distrib))
    inode=minloc(cost_node)
    call random_number(myrand)
    igroup=nint(myrand*nw)
   igroup=min(nw,max(1,igroup))
    if(not_distrib(igroup)) then
      cost_node(inode(1))=cost_node(inode(1))+mywork(igroup)
      not_distrib(igroup)=.false.
      cur_plan(igroup)=inode(1)
    end if
  end do

  mycost=(maxval(cost_node)-minval(cost_node))/minval(cost_node)
  if(mycost<lowest) then
    lowest=mycost
    low_plan=cur_plan
  end if
end do



#ifdef SEP_MPI
end if
call MPI_Bcast(low_plan,size(low_plan),MPI_INTEGER, 0, MPI_COMM_WORLD,ierr)
#endif
low_plan=low_plan-1
down%iw_thread=low_plan

!Figure out how many groups local iw is holding
down%nw_local=0
do i=1,size(low_plan)
  if(low_plan(i)==run%m_ith) down%nw_local=down%nw_local+1
end do

!Allocate the number of zs
allocate(var%z_samp(down%nw_local))
call add_mem(size(var%z_samp)/2)

j=0
allocate(down%iw_first(down%nw_local))
allocate(down%wts(down%nws,down%nw_local))
allocate(down%freqs(down%nws,down%nw_local))
allocate(down%w_local(down%nws))
allocate(down%wt_local(down%nws))
call add_mem(down%nw_local/2+down%nws*1+down%nws*down%nw_local*1)

   call from_param("nlagrange_pts",npts,5)


do i=1,size(low_plan)
  if(low_plan(i)==run%m_ith) then
    j=j+1
    down%iw_first(j)=(ifreq(i)-1)*down%nws+1
    do iww=1,down%nws
       down%freqs(iww,j)=freq_list(iww,ifreq(i))
       down%wts(iww,j)=wt_list(iww,ifreq(i))
    end do
    allocate(var%z_samp(j)%dz(nz(i)))
    allocate(var%z_samp(j)%status(nz(i)))
    allocate(var%z_samp(j)%zloc(nz(i)))
    allocate(var%z_samp(j)%iloc(npts,down%rsep%n(5)))
    allocate(var%z_samp(j)%frac(npts,down%rsep%n(5)))
    call add_mem(nint(nz(i)*1.5)+npts*down%rsep%n(5))
    var%z_samp(j)%dz(1:nz(i))=dz(1:nz(i),i)
    var%z_samp(j)%zloc(1)=down%rsep%o(5)
     do m=2,nz(i)
       var%z_samp(j)%zloc(m)=var%z_samp(j)%zloc(m-1)+var%z_samp(j)%dz(m-1)
     end do
   call lagrange_init(var%z_samp(j)%iloc,var%z_samp(j)%frac,var%z_samp(j)%zloc,&
    down%rsep%n(5),down%rsep%o(5),down%rsep%d(5))
  end if
end do


   call from_param("nz_mem",nz_mem,6)
   call from_param("nlagrange_pts",npts,5)
  if(run%mem_save>1) nz_mem=0
 
   allocate(var%filled(down%rsep%n(5)))
    call add_mem(down%rsep%n(5)/2)

   !in core z buffer
   if(nz_mem >0) then
     allocate(var%v_i_buf(down%dsep%n(1),down%dsep%n(2),down%dsep%n(3),down%dsep%n(4),nz_mem))
     allocate(var%mem_buf(nz_mem))
      call add_mem(size(var%mem_buf)/2+size(var%v_i_buf))
     var%nmem_buf=nz_mem
   else
     var%nmem_buf=0
     allocate(var%mem_buf(1))
      call add_mem(size(var%mem_buf)/2)
   end if


   !on disk z buffer
   if(npts-nz_mem >0) then
     var%ndisk_buf=npts-nz_mem
     var%use_disk_buf=.true.
     allocate(var%disk_buf(npts-nz_mem))
      call add_mem(size(var%disk_buf)/2)
   else
     var%NDISK_BUf=0
     var%use_disk_buf=.false.
     allocate(var%disk_buf(1))
      call add_mem(size(var%disk_buf)/2)
   end if


call del_mem(size(dz)+nw*6+size(cost_node))
deallocate(dz,nz,not_distrib,low_plan,cur_plan,mywork,cost_node)

distribute_var_freq_slow=T
end function

subroutine get_mid_freq_list(ifreq,freq_list,flist)
integer :: ifreq(:),iw
real :: freq_list(:,:),flist(:)

do iw=1,size(ifreq)
  flist(iw)=sum(freq_list(:,ifreq(iw)))/size(freq_list,1)
end do

end subroutine



!-------------------------------------------------------------
!
! Calculate the amount of work required for a given frequency
!
!
!


integer function calc_work(freq,min_dz,max_dz,max_cycle,sample_per,dz)
integer :: sample_per
logical :: slowness
integer :: n(3)
integer :: ipos,ithread,max_cycle
real :: d(3),freq,min_z,max_z,nax_dz,dv(3),max_Dz
real :: o(3),z,zstep,ov(3)
real :: dz(:)
integer i,npts,inew
real :: sample_hold(size(dz,1))
logical :: changed
integer :: iz
real :: fb,fe,min_dz,myf,cut,count
integer :: ib,ie,j,jb,je,ifreq,ierr,nv(3)
logical :: cutoff

nv=(/down%S%s_s%struct%n(1),down%S%s_s%struct%n(2),down%rsep%n(5)/)
ov=(/down%S%s_s%struct%o(1),down%S%s_s%struct%o(2),down%rsep%o(5)/)
dv=(/down%S%s_s%struct%d(1),down%S%s_s%struct%d(2),down%rsep%d(5)/)
min_z=down%rsep%o(5)
max_z=down%rsep%o(5) +(down%rsep%n(5)-1)*down%rsep%d(5)


n=nv
d=dv
o=ov


myf=freq

z=min_z;npts=1;sample_hold(npts)=z
cutoff=.false.
count=0
do while(z< max_z .and. .not. cutoff)
  ipos=min(n(3),max(1,floor((z-o(3))/d(3)+1.5)))
  !GIVEN THE CURRENT VELOCITY DEPTH LOCATION, CHECK TO SEE HOW
  !LARGE IN DEPTH THIS FREQUENCY VALUE AND ARE SAMPLING
  zstep=(1./max_slow(ipos)/myf)/sample_per
!  if(run%m_ith==0) write(0,*) "CHECK ZSTEP 1",zstep

  !check to see if there is a lower velocity with our given depth step
  inew=min(n(3),max(1,floor((z+zstep-o(3))/d(3)+1.5)))
  zstep=max(min_dz,(minval(1./max_slow(ipos:inew))/myf)/sample_per)
!  if(run%m_ith==0) write(0,*) "CHECK ZSTEP 2",zstep
  zstep=min(max_dz,zstep)
  z=z+zstep
  npts=npts+1;sample_hold(npts)=z
  count=count + freq*maxval(max_slow(ipos:inew))* zstep
  if(count>max_cycle) cutoff=.true.
!  if(run%m_ith==0) write(0,*) "CHECK ZSTEP 3",zstep,z,max_z,cutoff
end do


!CALCULATE DZ SAMPLING
do i=1,npts-1
  dz(i)=sample_hold(i+1)-sample_hold(i)
end do

calc_work=npts


end function calc_work

!------------------------------------------
!
!  Reduce the h-domain as we go down in depth
!
!
!

logical function init_reduce_h()
logical :: reduce_h
integer ::i,nmax,ifreq
logical :: rite_sample
integer, external :: getch
real :: hslope,h0,hmax,z0
init_reduce_h=F
 allocate(igood_len(230))
  call add_mem(size(igood_len))
igood_len=(/ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,&
     18, 20, 21, 22, 24, 26, 28, 30, 33, 35, 36, 39, 40, 42, 44,&
     45, 48, 52, 55, 56, 60, 63, 65, 66, 70, 72, 77, 78, 80, 84,&
     88, 90, 91, 99, 104, 105, 110, 112, 117, 120, 126, 130, 132,&
    140, 143, 144, 154, 156, 165, 168, 176, 180, 182, 195, 198, 208,&
    210, 220, 231, 234, 240, 252, 260, 264, 273, 280, 286, 308, 312,&
    315, 330, 336, 360, 364, 385, 390, 396, 420, 429, 440, 455, 462,&
    468, 495, 504, 520, 528, 546, 560, 572, 585, 616, 624, 630, 660,&
    693, 715, 720, 728, 770, 780, 792, 819, 840, 858, 880, 910, 924,&
    936, 990, 1001, 1008, 1040, 1092, 1144, 1155, 1170, 1232, 1260,&
   1287, 1320, 1365, 1386, 1430, 1456, 1540, 1560, 1584, 1638, 1680,&
   1716, 1820, 1848, 1872, 1980, 2002, 2145, 2184, 2288, 2310, 2340,&
   2520, 2574, 2640, 2730, 2772, 2860, 3003, 3080, 3120, 3276, 3432,&
   3465, 3640, 3696, 3960, 4004, 4095, 4290, 4368, 4620, 4680, 5005,&
   5040, 5148, 5460, 5544, 5720, 6006, 6160, 6435, 6552, 6864, 6930,&
   7280, 7920, 8008, 8190, 8580, 9009, 9240, 9360, 10010, 10296,&
  10920, 11088, 11440, 12012, 12870, 13104, 13860, 15015, 16016,&
  16380, 17160, 18018, 18480, 20020, 20592, 21840, 24024, 25740,&
  27720, 30030, 32760, 34320, 36036, 40040, 45045, 48048, 51480,&
  55440, 60060, 65520, 72072, 80080, 90090, 102960, 120120, 144144,&
 180180, 240240, 360360, 720720/)

call from_param("reduce_h",var%reduce_h,.false.)

!MAYBE WE CAN GET RID OF THIS????
if(var%reduce_h .and. run%mem_save>1) then
  write(0,*)  "can not reduce h and do memory save"
end if

call from_param("reduce_h_h0",h0,max(&
  max(abs(down%dsep%o(3)),abs(down%dsep%o(3)+down%dsep%d(3)*(down%dsep%n(3)-1))),&
  max(abs(down%dsep%o(4)),abs(down%dsep%o(4)+down%dsep%d(4)*(down%dsep%n(4)-1)))))
call from_param("reduce_h_z0",z0,down%rsep%o(5)+down%rsep%d(5)*(down%rsep%n(5)-1))
call from_param("reduce_h_slope",hslope,0.)
nmax=1;
do i=1,size(down%iw_first)
 nmax= max(size(var%z_samp(i)%dz),nmax)
end do

allocate(var%ihyb_beg(nmax))
allocate(var%ihxb_beg(nmax))
allocate(var%ihyb_end(nmax))
allocate(var%ihxb_end(nmax))
allocate(ihxbeg(nmax,size(down%iw_first)))
allocate(ihxend(nmax,size(down%iw_first)))
allocate(ihybeg(nmax,size(down%iw_first)))
allocate(ihyend(nmax,size(down%iw_first)))
allocate(zdepth(nmax,size(down%iw_first)))
call add_mem(nmax*4+nmax*4*size(down%iw_first))
ihxbeg=1
ihybeg=1
ihxend=down%dsep%n(3)
ihyend=down%dsep%n(4)

zdepth=0
 do ifreq=1,size(down%iw_first)
  do i=1,size(var%z_samp(ifreq)%dz)
    if(i>1) zdepth(i,ifreq)=zdepth(i-1,ifreq)+var%z_samp(ifreq)%dz(i-1)
if (var%reduce_h) then
    if(zdepth(i,ifreq) > z0) then
      hmax=h0+hslope*(zdepth(i,ifreq)-z0)
      if(hmax < 0) then
        write(0,*) "h dist went zero change hslope, z0, and/or h0"
          call erexit("")
       end if
      if(down%dsep%n(4)>1) then
        call set_limits(down%dsep,4,ihybeg(i,ifreq),ihyend(i,ifreq),hmax)
      end if
      if(down%dsep%n(3)>1) then
        call set_limits(down%dsep,3,ihxbeg(i,ifreq),ihxend(i,ifreq),hmax)
end if
      end if
    end if
  end do
 end do


init_reduce_h=T
end function



!-------------------------------------------------------------
!
!  Find the limits of the h-domain for a given depth
!
!

subroutine set_limits(struct,iax,ibeg,iend,max_dist)
integer :: ibeg,iend,ileft,iright
real :: mymax,max_dist
type(axis) :: ax
type(sep3d) :: struct
integer :: inext,i,iax
  ax%n=struct%n(iax)
  ax%o=struct%o(iax)
  ax%d=struct%d(iax)


  mymax=max(abs(ax%o),abs(ax%o+ax%d*(ax%n-1)))
  ibeg=1
  iend=ax%n
  ileft=0
  iright=0
  if(mymax > max_dist) then
    if(-ax%o > max_dist) then
      ileft=(-ax%o-max_dist)/ax%d
    end if
    if(ax%o > max_dist) then
      ileft=(ax%o-max_dist)/ax%d
    end if
    if(-(ax%o+ax%d*(ax%n-1)) > max_dist) then
      iright=-((ax%o+ax%d*(ax%n-1))-max_dist)/ax%d
    end if
    if((ax%o+ax%d*(ax%n-1)) > max_dist) then
      iright=((ax%o+ax%d*(ax%n-1))-max_dist)/ax%d
    end if
    inext=next_fast(ax%n-ileft-iright)
    if(inext < ax%n) then
      do while(iend-ibeg+1 /= inext)
        if(ileft > iright) then
          ileft=ileft-1
          ibeg=ibeg+1
        else
          iright=iright-1
          iend=iend-1
        end if
      end do
     end if
   end if
end subroutine set_limits
!-------------------------------------------------------------
!
!  Find the next good fft length
!
!

integer function next_fast(nlow)
integer i,nlow
logical :: found
  i=1; found=.false.
  do while(i < size(igood_len) .and. .not. found)
    if(igood_len(i)< nlow) then
     i=i+1
    else
     found=.true.
    end if
  end do
  next_fast=igood_len(i)
end function
!-------------------------------------------------------------
!
!  Figure out the variable stepping for a given frequency
!
!


  subroutine setup_var_block(iws,izs)
    integer :: iws,izs
    integer :: j,k,l,i,iz
    logical :: found
    if(down%order(1:3)/="VAR")  then
       var%dzb=down%rsep%d(5)
       var%ihxb_beg=1
       var%ihyb_beg=1
       var%ihxb_end=down%dsep%n(3)
       var%ihyb_end=down%dsep%n(4)
    else
      j=(izs-1)*down%nzs
      k=min(size(var%z_samp(iws)%dz),down%nzs*izs)
      l=k-j
      var%dzb(1:l)=var%z_samp(iws)%dz(j+1:k)
      do iz=1,l
       i=1; found=.false.
       do while (i< size(zdepth,1) .and. .not. found )
        if(var%z_samp(iws)%zloc(iz+j) < zdepth(i,iws)) then
          found=.true.
        else
         i=i+1
        end if
       end do
        var%ihxb_beg=ihxbeg(i,iws)
        var%ihxb_end=ihxend(i,iws)
        var%ihyb_beg=ihybeg(i,iws)
        var%ihyb_end=ihyend(i,iws)
      end do
    end if
    down%w_local=down%freqs(:,iws)
    down%wt_local=down%wts(:,iws)
  end subroutine setup_var_block

!--------------------------------------------------------------
!
!
!
!


logical function  store_distrib_pattern()
  integer :: ierr

 store_distrib_pattern=F


if(run%m_ith==0) then
   if(1==getch("distrib_pattern","d",ierr)) then
    call to_history("n1",size(down%iw_thread),"distrib_pattern")
    ierr=srite("distrib_pattern",real(down%iw_thread),size(down%iw_thread)*4)
  end if
end if
 store_distrib_pattern=T

end function


!--------------------------------------------------------------
!
! Initialize the variable stepping buffers
!
!

  subroutine  var_samp_iw(iwb)
    integer :: iwb
    var%filled=.false.
    var%mem_buf=0
    var%disk_buf=0
    var%z_samp(iwb)%status=0
  end subroutine var_samp_iw

end module
