module wei_cam
  !! Wave-Equation Migration
  !! Common-azimuth

  use wei_util
  use wei_external
  use wei_down
  use wei_run
  use wei_obj_S
  use wei_kmap
  use wei_fft
  !---------------
  implicit none

  real,private                            :: dkxko, cdkxko,idkzko,rdkzko
  integer     ,private                    :: nkxko
  real                            ::  kxko
  real,    dimension(:), pointer,private :: ckzko,exp_kzko,ikzko
  complex, dimension(:), pointer,private ::  kzko, tkzko
  complex, pointer, private :: w_extrap_1(:,:,:,:,:),w_extrap_2(:,:,:,:,:)
  real, dimension(:) , pointer,private :: khy(:,:)
contains
  !----------------------------------------------------------------

  subroutine cam_dsr_init(max,n)
    real,    intent(in) :: max
    integer, intent(in) :: n
    integer             :: er
    integer             :: i,m
    real                :: kxko,kz

    allocate( kzko(n),stat=er)
    call add_mem(size(kzko))
    if(er/=0) call seperr('cant allocate  kzko')
    allocate(tkzko(n),stat=er)
    call add_mem(size(tkzko))
    if(er/=0) call seperr('cant allocate tkzko')
    allocate(khy(down%wsep%n(1),down%wsep%n(3)))
    allocate(exp_kzko(n),ikzko(n))
    dkxko = max/(n-1)
    nkxko = n
    m=n/max
   

    !!                    ____________
    !!   kz         1    /     | kx |2          kx
    !! i -- dz = -i -   /  1 - | -- |  dz   0 < -- < 1 
    !!   ko         2 \/       | ko |           ko
    !!
    do i=1,m
       kxko    = 0. +     1.0 * (i  -1)/(  m-1) !! kx/ko=0...1
       kzko(i) = - (0,+1) * sqrt(1-kxko**2)
    end do
    idkzko=1./(n-1)
    do i=1,n
       kxko    = 0. +     1.0 * (i  -1)/(  n-1) !! kx/ko=0...1
       kz= sqrt(1.-kxko)
       ikzko(i)=kz
    end do

    !!                    ____________
    !!   kz         1    / | kx |2              kx
    !! i -- dz = -  -   /  | -- | - 1  dz   1 < -- < max 
    !!   ko         2 \/   | ko |               ko
    !!
    do i=m+1,n
       kxko    = 1. + (max-1) * (i-m-1)/(n-m-1) !! kx/ko=1...max
       kzko(i) = - (+1,0) * sqrt(kxko**2-1)
    end do

    rdkzko=(max-1.)/(n-1)
    do i=1,n
       kxko    = 1. +     (max-1.) * (i  -1)/(  n-1) !! kx/ko=0...1
       kz= sqrt(kxko-1.)
!       exp_kzko(i)=exp(kz)
       exp_kzko(i)=kz
    end do


    kzko = .5 * kzko
  end subroutine cam_dsr_init
  !----------------------------------------------------------------
!!$=head1 NAME
!!$
!!$weicam_init -initialize  common azimuth migration
!!$
!!$=head1 SYNOPSIS
!!$
!!$ierr= weicam_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$Common Azimuth Migration Initialization operator
!!$
!!$Grabs kxko=2. and nkxko=2001 from par
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weicamo> L<weiwem_init>
!!$
!!$=cut
!!$
  subroutine comaz_init(max,n)
    integer, intent(in) :: n
    real,    intent(in) :: max
    integer             :: er
    integer             :: i,m
    real                :: kxko

    allocate(ckzko(n),stat=er)
    call add_mem(size(ckzko))
    if(er/=0) call seperr('cant allocate ckzko')


    cdkxko = max/(n-1)
    m=n/max

    do i=1,m
       kxko     = 1. * (i-1)/(m-1) !! kx/ko=0...1
       ckzko(i) = sqrt(1-kxko**2)
    end do
    do i=m+1,n
       ckzko(i) = 0
    end do

!    ckzko(n) = ckzko(n-1) !! avoid kzs=0 or kzr=0
    ckzko = .5 * ckzko
  end subroutine comaz_init
  !----------------------------------------------------------------



  integer function weicam_gsp_init() result(st)
    integer :: er,n
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam_init'
    if(run%debug) call in(unit)
#endif

     n=size(tmp_on,5)
     if(run%fk_extra .ne. 2) call seperr("expecting  fk_extra ==2 for gsp")
     w_extrap_1=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
     n=n-run%o_nth
     w_extrap_2=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
     OK=weicam_init()
     down%maxslow=.true.

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weicam_gsp_init



!!$=head1 NAME
  integer function weicam_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam_init'
    if(run%debug) call in(unit)
#endif

    !! init SSR
    call from_param("nkxko",nkxko,2001)
    call from_param( "kxko", kxko,2.)
    call cam_dsr_init(kxko,nkxko)
    call  comaz_init(kxko,nkxko)

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weicam_init



!!$=head1 NAME
!!$
!!$weicam - one depth step of common azimuth migration
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weicam(iws,izs,ifk,ith,wfld)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item iws - integer
!!$
!!$  current frequency slice
!!$
!!$=item izs - integer
!!$
!!$  current depth slice
!!$
!!$=item ifk - integer
!!$
!!$  current fk operation
!!$
!!$=item ith - integer
!!$
!!$  current thread
!!$
!!$=item wfld - complex(:,:,:,:,:)
!!$
!!$  wavefield
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$Common Azimuth Migration  operator
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<wei_caminit> L<weiwem>
!!$
!!$=cut
  !----------------------------------------------------------------
  integer function weicam_gsp(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) ::iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    :: w,kos,kor
    real    :: inv_norm_kos,inv_norm_kor,inv_norm_ko,inv_norm_ko2
    real    :: r_inv_norm_ko,i_inv_norm_ko
    integer :: imx,imy,ihx,ihy 
    integer :: is,ir,i,sign
    complex :: ikz
    real    :: kzs,kzr,k_y
    real    :: ksy,kry,k,j,kmy
    real    :: kss,krr,ko,fract,iko2,cract,inv_slow_cb
    real,pointer :: kx(:,:),ky(:,:)
    !NEW
    real :: slow_0,epsi,slow_0_sq,w_data,kow,k_sq_comaz,dz
    real :: cconst,cconst1
    real :: a0,dz_w,d,myr,myi
    real :: inv_slow_0,c1,slow_min,slow_max,kow_sq,inv_gamma
    real :: det,gamma
    real  :: smin,smax
!BEGIN ORDER 2
    real :: inv_slow_0_cb
    complex :: c_etak2,c_shift2,rb,c_gamma_cb,cinv_gamma_cb,inv_gamma_sq,inv_gamma_cb
!END ORDER2
    complex :: cphase,c_gamma,c_det,c_kow,c_kow_sq,inv_gmma
    complex :: c_shift,cinv_gamma,c_etak1
    complex :: ra,c_shift1,s
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam'
    if(run%depar) call in(unit)
#endif

!! set constants
    !FIGURE OUT WHAT SOURCE WE ARE USING
    if(slo%th(ith)%source) then
      sign=1
      slow_0=slo%sz_s(izs,ifk)
      kx=>kmap%krx
    else
      slow_0=slo%sz_r(izs,ifk)
      sign=-1
      kx=>kmap%krx
    end if

 
    epsi=epsilon(1.)
    w= down%w_local(iws)
    kos=2* slow_0 * w;kor=2* slow_0 * w;ko = 2* slow_0 * w
    inv_norm_kos=1./(kos*cdkxko);inv_norm_kor=1./(kor*cdkxko)
    inv_norm_ko2=1./(ko*ko)
    iko2=1./ko**2
    r_inv_norm_ko=1./(rdkzko)
    i_inv_norm_ko=1./(idkzko)


  slow_0_sq=slow_0*slow_0
  inv_slow_0=1.0/slow_0
  inv_slow_0_cb=1.0/slow_0**3
  w_data=w
  dz_w=var%dzb(izs)*w_data

  call weiift(ith,izs,ith,w_extrap_1)
  call weiift(ith,izs,ith,w_extrap_2)
  ! Some GSP stuff
  ! constants of the (complex) contour deformations
  ! 1 refers to the contour deformation related to 1/gamma
  cconst=0.00009 /minval(down%wsep%d(1:3))
  cconst1=2.0*cconst
  c1=cconst1*cconst1
  ! slow_min degining of the contour deformation
  ! slow_max end of the contour deformation
  slow_min=slow_0-cconst1*.995  !FACTOR TO COUNTER ROUNDING
  slow_max=slow_0+cconst1*.995

  !CALCULATE VARIABLE BOUNDARIES GIVEN OFFSET REDUCTION
  if(var%ihyb_beg(izs) >1) wfld(:,:,:,1:var%ihxb_beg(izs)-1,iws)=0.
  if(var%ihyb_end(izs) < down%wsep%n(4)) wfld(:,:,:,var%ihyb_end(izs)+1:down%wsep%n(4),iws)=0.
  if(var%ihxb_beg(izs) >1) wfld(:,:,1:var%ihxb_beg(izs)-1,var%ihyb_beg(izs):var%ihyb_end(izs),iws)=0.
  if(var%ihxb_end(izs) < down%wsep%n(3))  &
  wfld(:,:,var%ihxb_end(izs)+1:down%wsep%n(3),var%ihyb_beg(izs):var%ihyb_end(izs),iws)=0.

  ! Constants for spatial conrrection terms
  ! as given by the generalized-screen expansion
  a0=-dz_w
  slow_max=slow_0+cconst1
  s=cmplx(0,-w)
  dz=var%dzb(izs)
  smax=(slow_max*w_data*2)**2
  smin=(slow_min*w_data*2)**2
  w_extrap_1(:,:,:,:,ith)=s/2.*dz*w_extrap_1(:,:,:,:,ith)
  w_extrap_2(:,:,:,:,ith)=-s/8.*dz*w_extrap_2(:,:,:,:,ith)
   

    if(down%opsign==goB2T) then
       tkzko=conjg(kzko)
       fract=.5*var%dzb(izs)*ko
       cract=slow_0
    else
       tkzko=kzko
       fract=-.5*var%dzb(izs)*ko
       cract=slow_0
    end if


    !FIGURE OUT WHAT SOURCE WE ARE USING
    if(slo%th(ith)%source) then
      sign=1
      kos=2* slo%sz_s(izs,ifk) * w
      kor=2* slo%sz_s(izs,ifk) * w
      ko = 2* slo%sz_s(izs,ifk) * w
      kx=>kmap%krx
    else
      sign=-1
      kos=2* slo%sz_r(izs,ifk) * w
      kor=2* slo%sz_r(izs,ifk) * w

      ko = 2* slo%sz_r(izs,ifk) * w
      kx=>kmap%ksx
    end if

!  call calc_khy(kos,kor,izs)
  call calc_khy(kos,inv_norm_kos,kor,inv_norm_kor,izs)
   

if(ko >0.) then 
  do ihx=var%ihxb_beg(izs),var%ihxb_end(izs) 
    do imy=1,down%wsep%n(2) 
      kmy=kmap%kmy(imy)
      do imx=1,down%wsep%n(1) 
        if (abs(wfld(imx,imy,ihx,1,iws))>0.0) then
              ! evaluate DSR with common-azimuth k_hy
              k_y=kmy+sign*kmy*khy(imx,ihx)
              k_sq_comaz=kx(imx,ihx)**2+k_y**2
                 k = iko2*k_sq_comaz
                 if (k<1.) then
                    i= min(nkxko,max(1, int(1 + k*i_inv_norm_ko)))
                    gamma=(ikzko(i)*cract )
                    inv_gamma=1./max(gamma,.000001)
                    j=fract*ikzko(i)
                    c_shift=cmplx(cos(j),sin(j))
                    c_etak1=inv_gamma-inv_slow_0
                    c_etak2=(inv_gamma**3-inv_slow_0_cb)
                 else
                    i=max(1, min(nkxko,int(1 + (k-1.)*r_inv_norm_ko)))
                    inv_gamma=1.0/max(.000001,(cract*exp_kzko(i)))
                    j=fract*exp_kzko(i) 
                    c_shift=cmplx(exp(j), 0.0)
                    c_etak1=cmplx(inv_gamma, -inv_slow_0)
                    c_etak2=cmplx(inv_gamma**3,-inv_slow_0_cb)
                 end if
                    
                   

!!BEGIN ORDER 1
!                 ra = (w_extrap_1(imx,imy,ihx,1,ith)*c_etak1) &
!                      / wfld(imx,imy,ihx,1,iws)
!END ORDER 1
!BEGIN ORDER 2
                ra = ((w_extrap_1(imx,imy,ihx,1,ith)*c_etak1) &
                      + (w_extrap_2(imx,imy,ihx,1,ith)*c_etak2)) &
                      / wfld(imx,imy,ihx,1,iws)
!END ORDER2
                 myr=real(ra)
                 myi=aimag(ra)
                 ra=1+myr/cmplx(1,myi)
                 c_shift1=cexp(cmplx(0.0,myi))*ra/cabs(ra)
                 wfld(imx,imy,ihx,1,iws) = &
                 wfld(imx,imy,ihx,1,iws)*c_shift*c_shift1
              else
                 wfld(imx,imy,ihx,1,iws) = cmplx(0.0,0.0)
              end if

              ! GSP stuff finishes here
           end do 
        end do 
     end do 
  end if 

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weicam_gsp
  !----------------------------------------------------------------
  integer function weicam(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) ::iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    :: w,kos,kor
    real    :: inv_norm_kos,inv_norm_kor,inv_norm_ko,inv_norm_ko2
    real    :: r_inv_norm_ko,i_inv_norm_ko
    integer :: imx,imy,ihx,ihy 
    integer :: is,ir,i,sign
    complex :: ikz
    real    :: kzs,kzr,k_y
    real    :: ksy,kry,k,j,kmy
    real    :: kss,krr,ko,fract,iko2
    real,pointer :: kx(:,:),ky(:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam'
    if(run%depar) call in(unit)
#endif
    w= down%w_local(iws)

    !FIGURE OUT WHAT SOURCE WE ARE USING


    if(slo%th(ith)%source) then
      sign=-1
      ko = 2* slo%sz_s(izs,ifk) * w
      kos=2* slo%sz_s(izs,ifk) * w
      kor=2* slo%sz_s(izs,ifk) * w
      kx=>kmap%ksx
    else
      sign=1
      ko = 2* slo%sz_r(izs,ifk) * w
      kos=2* slo%sz_r(izs,ifk) * w
      kor=2* slo%sz_r(izs,ifk) * w
      kx=>kmap%krx
    end if

    inv_norm_kos=1./(kos*cdkxko)
    inv_norm_kor=1./(kor*cdkxko)
    iko2=1./ko**2
    inv_norm_ko2=1./(ko*ko)
    r_inv_norm_ko=1./(rdkzko)
    i_inv_norm_ko=1./(idkzko)

    if(down%opsign==goB2T) then
       tkzko=conjg(kzko)
       fract=.5*var%dzb(izs)*ko
    else
       tkzko=kzko
       fract=-.5*var%dzb(izs)*ko
    end if



    !CALCULATE VARIABLE BOUNDARIES GIVEN OFFSET REDUCTION
    if(var%ihyb_beg(izs) >1) wfld(:,:,:,1:var%ihxb_beg(izs)-1,iws)=0.
    if(var%ihyb_end(izs) < down%wsep%n(4)) wfld(:,:,:,var%ihyb_end(izs)+1:down%wsep%n(4),iws)=0.
    if(var%ihxb_beg(izs) >1) wfld(:,:,1:var%ihxb_beg(izs)-1,var%ihyb_beg(izs):var%ihyb_end(izs),iws)=0.
    if(var%ihxb_end(izs) < down%wsep%n(3))  &
    wfld(:,:,var%ihxb_end(izs)+1:down%wsep%n(3),var%ihyb_beg(izs):var%ihyb_end(izs),iws)=0.

!  call calc_khy(kos,kor,izs)
  call calc_khy(kos,inv_norm_kos,kor,inv_norm_kor,izs)
   
if(ko >0.) then
    do ihy=var%ihyb_beg(izs),var%ihyb_end(izs)
       do ihx=var%ihxb_beg(izs),var%ihxb_end(izs)
          do imy=1,down%wsep%n(2)
             kmy=kmap%kmy(imy)
             do imx=1,down%wsep%n(1)
                k_y =kmy+ sign*khy(imx,ihx)*kmy
                k = iko2*(kx(imx,ihx)**2 + k_y**2)
                if(k < 1.) then
                    i= min(nkxko,max(1, int(1 + k*i_inv_norm_ko)))
                    j=fract*ikzko(i)
                   wfld(imx,imy,ihx,ihy,iws) = wfld(imx,imy,ihx,ihy,iws) * &
                     cmplx(cos(j),sin(j))
                else
                   i=max(1, min(nkxko,int(1 + (k-1.)*r_inv_norm_ko)))
                   j=fract*exp_kzko(i)
                   wfld(imx,imy,ihx,ihy,iws) = &
                   wfld(imx,imy,ihx,ihy,iws) * exp(j)
                end if

             end do
          end do
       end do
    end do
end if

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weicam
  !----------------------------------------------------------------

subroutine calc_khy(kos,inv_norm_kos,kor,inv_norm_kor,izs)
integer :: is,ir,ihx,imx,izs
real :: kzs,kzr,kos,kor,inv_norm_kos,inv_norm_kor
 do ihx=var%ihxb_beg(izs),var%ihxb_end(izs)
   do imx=1,down%wsep%n(1)
        is = min(int(1 + abs(kmap%ksx(imx,ihx)*inv_norm_kos)) , nkxko)
        ir = min(int(1 + abs(kmap%krx(imx,ihx)*inv_norm_kor)) , nkxko)
        kzs = kos*ckzko(is)
        kzr = kor*ckzko(ir)
        if((kzs > 0.) .and. (kzr >0.))then 
           khy(imx,ihx) =(kzr-kzs)/(kzs+kzr)
        else
           khy(imx,ihx) = 0.
       end if
   end do
 end do
end subroutine



end module wei_cam
