!HANDLE FFTS
module wei_fft
  !! FFTW
  !! used for data(mx,my,hx,hy,w)

!IF WE AREN'T USING FFTW FAKE IT WITH SU'S PFFT
#ifndef FFTW
use fftw_sep_fake_mod
#endif


  use sep_timers_mod
  use wei_down
  use wei_run
  use wei_flags
  use wei_attr
  !---------------
  implicit none
  integer, private, save :: ndim_f
  integer(kind=dbl), private,save :: planF(:,:,:) !! forward FFT
  integer(kind=dbl), private,save :: planI(:,:,:) !! inverse FFT
  integer, private,save ::  ndim_p(:,:,:)
  real,      private :: scale
  logical,private,save :: fftmeasure
  integer, pointer :: nplans(:),cur_plan(:)

  !! temp  arrays
  complex, dimension(:,:),         pointer, private ::   temp1,temp2

  !! shift arrays
  complex, dimension(:,:),           pointer, private :: sftF
  complex, dimension(:,:),           pointer, private :: sftI

contains
  !----------------------------------------------------------------
  !INITIALIZE THE FFT
  subroutine weifft_init()
    integer, dimension(5) :: ni
    real    :: arg
    integer :: ihx,ihy,i
    integer :: er,max_p
#ifdef DEBUG
    character(len=128), parameter :: unit='weifft_init'
    if(run%debug) call in(unit)
#endif

    !WHETHER OR NOT TO CALCULATE THE SPEED OF VARIOUS FFT METHODS
    call from_param("fftmeasure",fftmeasure,.false.)

    !! FFT
    call compress_axes(down%wsep%n,ndim_f,ni)
    scale = 1./sqrt(1.*product(ni))

    allocate(temp1(ni(1)*ni(2),run%o_nth),temp2(ni(3)*ni(4),run%o_nth))
    call add_mem(size(temp1)+size(temp2))

    

    call from_param("max_p",max_p,4)
    allocate(planF(2, max_p,run%o_nth))
      call add_mem(size(planF))
    allocate(planI(2, max_p,run%o_nth))
      call add_mem(size(planI))
    allocate(ndim_p(5,max_p,run%o_nth))
      call add_mem(size(ndim_p))
    allocate(nplans(run%o_nth))
      call add_mem(size(nplans))
    allocate(cur_plan(run%o_nth))
      call add_mem(size(cur_plan))
    nplans=0
    cur_plan=0

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weifft_init

 !GIVEN A DATASET DIMENSION
 !FIND IF WE HAVE ALREADY CREATED AN FFTW PLAN FOR IT
 !IF NOT CREATE ONE
 integer function find_plan(ni,ith)
   integer  :: i,ni(5),ith
   logical :: found

     i=1
     found=.false.
    do while(i <= nplans(ith) .and. .not. found) 
      if(all(ni == ndim_p(:,i,ith))) then
       found=.true.
      else
       i=i+1
     end if
    end do
     if(.not. found) then
       call add_plan(ni,ith)
       i=cur_plan(ith)
       found=.true.
     end if
     find_plan=i
 end function find_plan

 subroutine compress_axes(nin,ndim,nout)
 integer :: nin(:),nout(:),ndim,i
    ndim=0
    nout=1
    do i=1,4
      if(nin(i).ne. 1) then
         ndim=ndim+1
         nout(ndim)=nin(i)
      end if
    end do
 end subroutine

 !ADD A NEW PLAN FOR A NEW DIMENSION SIZE
 subroutine add_plan(ni,ith)
  integer, dimension(5) :: ni,ni2
  integer :: ith,ndim,i
    if(nplans(ith)<size(cur_plan)) then
      cur_plan(ith)=cur_plan(ith)+1
      nplans(ith)=nplans(ith)+1
    else 
     if(cur_plan(ith)==size(cur_plan)) then
       cur_plan(ith)=1
     else
       cur_plan(ith)=cur_plan(ith)+1
     end if 
      call fftwnd_f77_destroy_plan(planF(1,cur_plan(ith),ith))
      call fftwnd_f77_destroy_plan(planI(1,cur_plan(ith),ith))
      call fftwnd_f77_destroy_plan(planF(2,cur_plan(ith),ith))
      call fftwnd_f77_destroy_plan(planI(2,cur_plan(ith),ith))
    end if
     call compress_axes(ni,ndim,ni2)
    ndim_p(:,cur_plan(ith),ith)=(/ni(1:4),1/)

    if(fftmeasure) then
       !! init  forward plan
       call fftwnd_f77_create_plan(planF(1,cur_plan(ith),ith),min(2,ndim),ni2,&
         FFTW_FORWARD,fFTW_MEASURE +FFTW_USE_WISDOM+ FFTW_THREADSAFE+FFTW_IN_PLACE )
       !! init backward plan
       call fftwnd_f77_create_plan(planI(1,cur_plan(ith),ith),min(2,ndim),ni2,&
        FFTW_BACKWARD  , fFTW_MEASURE + FFTW_THREADSAFE +FFTW_USE_WISDOM+FFTW_IN_PLACE) 

       call fftwnd_f77_create_plan(planF(2,cur_plan(ith),ith),max(1,ndim-2),ni2(3),&
         FFTW_FORWARD,fFTW_MEASURE +FFTW_USE_WISDOM+ FFTW_THREADSAFE+FFTW_IN_PLACE )
       !! init backward plan
       call fftwnd_f77_create_plan(planI(2,cur_plan(ith),ith),max(1,ndim-2),ni2(3),&
        FFTW_BACKWARD  , fFTW_MEASURE + FFTW_THREADSAFE +FFTW_USE_WISDOM+FFTW_IN_PLACE) 
    else
       !! init  forward plan
       call fftwnd_f77_create_plan(planF(1,cur_plan(ith),ith),min(2,ndim),ni2,&
        FFTW_FORWARD ,  fFTW_ESTIMATE + fFTW_THREADSAFE + FFTW_USE_WISDOM+FFTW_IN_PLACE)
       !! init backward plan
       call fftwnd_f77_create_plan(planI(1,cur_plan(ith),ith),min(2,ndim),ni2,&
         FFTW_BACKWARD  ,fFTW_ESTIMATE + fFTW_THREADSAFE + FFTW_USE_WISDOM+FFTW_IN_PLACE)
       !! init  forward plan
       call fftwnd_f77_create_plan(planF(2,cur_plan(ith),ith),max(1,ndim-2),ni2(3),&
        FFTW_FORWARD ,  fFTW_ESTIMATE + fFTW_THREADSAFE + FFTW_USE_WISDOM+FFTW_IN_PLACE)
       !! init backward plan
       call fftwnd_f77_create_plan(planI(2,cur_plan(ith),ith),max(1,ndim-2),ni2(3),&
         FFTW_BACKWARD  ,fFTW_ESTIMATE + fFTW_THREADSAFE + FFTW_USE_WISDOM+FFTW_IN_PLACE)
    end if
  end subroutine add_plan 



  !----------------------------------------------------------------
  !! forward FFT
  subroutine weifft(iws,izs,ith,data)
    integer, intent(in) :: iws,ith,izs
    complex, dimension(:,:,:,:,:) :: data
    integer :: imx,imy,ndim
    integer :: iplan,i3,i1,n1,n2,i2,i4
    integer :: nd(4),nf(4)
    real    :: in123
#ifdef DEBUG
    character(len=128), parameter :: unit='weifft'
    if(run%depar) call in(unit)
#endif
     nd=(/size(data,1),size(data,2), var%ihxb_end(izs) -var%ihxb_beg(izs)+1, &
     var%ihyb_end(izs) -var%ihyb_beg(izs)+1/)
     iplan=find_plan((/nd(1),nd(2),nd(3),nd(4),1/),ith)
     call fft_2pass(planF(:,iplan,ith),iws,izs,ith,data,nd)



 end subroutine
 subroutine fft_2pass(plan,iws,izs,ith,data,nd)
  integer(kind=dbl)             :: plan(2)
  integer                       :: iws,izs,ith,nd(4),nf(4),ib,ie,ib2,ie2
  complex, dimension(:,:,:,:,:) :: data
  integer                       :: ndim,i1,i2,i3,i4
  real                          :: in123
  
   
     call compress_axes(nd,ndim,nf)
     in123=sqrt(1./product(nd(1:4)))
     if(nd(1) .ne. 1 ) then
       if(nd(2) .ne. 1) then !3-D DATA
!         do i4=var%ihyb_beg(izs),var%ihyb_end(izs)
!           do i3=var%ihxb_beg(izs),var%ihxb_end(izs)
!             temp1(:,ith)=reshape(data(:,:,:,:,iws),(/size(temp1(:,1))/))*in123
!             call fftwnd_f77_one(plan(1),temp1(:,ith),data(:,:,i3,i4,iws))
#ifdef SEP_TIMER
       if(run%m_ith==0) call start_timer_num(run%fftw_t)
#endif
             call fftwnd_f77(plan(1),nd(3)*nd(4),data(:,:,:,:,iws), 1,&
             nd(1)*nd(2),0,1,nd(1)*nd(2))
#ifdef SEP_TIMER
       if(run%m_ith==0) call stop_timer_num(run%fftw_t)
#endif
!           end do 
!         end do 
       else  !2-D x
         ib=var%ihxb_beg(izs)
         ie=var%ihxb_end(izs)
!           temp1(:,ith)=reshape(data(:,:,ib:ie,i4,iws),(/size(temp1(:,1))/))*in123
#ifdef SEP_TIMER
       if(run%m_ith==0) call start_timer_num(run%fftw_t)
#endif
           call fftwnd_f77_one(plan(1),data(:,:,:,:,iws),0)
#ifdef SEP_TIMER
       if(run%m_ith==0) call stop_timer_num(run%fftw_t)
#endif
       end if
     else !2-D y
       ib=var%ihyb_beg(izs)
       ie=var%ihyb_end(izs)
       temp1(:,ith)=reshape(data(:,:,:,ib:ie,iws),(/size(temp1(:,1))/))*in123
#ifdef SEP_TIMER
       if(run%m_ith==0) call start_timer_num(run%fftw_t)
#endif
       call fftwnd_f77_one(plan(1),temp1(:,ith),data(:,:,:,ib:ie,iws))
#ifdef SEP_TIMER
       if(run%m_ith==0) call stop_timer_num(run%fftw_t)
#endif
     end if

     if(ndim>2) then !IF WE HAVE 3-D DATA
       ib=var%ihxb_beg(izs)
       ie=var%ihxb_end(izs)
       ib2=var%ihyb_beg(izs)
       ie2=var%ihyb_end(izs)
!       do i2=1,nd(2)
!         do i1=1,nd(1)
!           temp2(:,ith)=reshape(data(i1,i2,ib:ie,ib2:ie2,iws),(/size(temp2(:,1))/))
#ifdef SEP_TIMER
       if(run%m_ith==0) call start_timer_num(run%fftw_t)
#endif
!           call fftwnd_f77_one(plan(2),temp2(:,ith),data(i1,i2,ib:ie,ib2:ie2,iws))
           call fftwnd_f77(plan(2),nd(1)*nd(2),data(:,:,:,:,iws),nd(1)*nd(2),&
             1,0,nd(1)*nd(2),1)
#ifdef SEP_TIMER
!       if(run%m_ith==0) call stop_timer_num(run%fftw_t)
#endif
!         end do
!       end do
     end if
     
           data(:,:,:,:,iws)= data(:,:,:,:,iws)*in123
       if(run%m_ith==0) call stop_timer_num(run%fftw_t)




#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine
  !----------------------------------------------------------------
  !! inverse FFT
  subroutine weiift(iws,izs,ith,data)
    integer, intent(in) :: iws, ith,izs
    complex, dimension(:,:,:,:,:), pointer :: data
    integer :: imx,imy,iplan,nd(4),i1,i2,i3,i4
    real :: in123
#ifdef DEBUG
    character(len=128), parameter :: unit='weiift'
    if(run%depar) call in(unit)
#endif
     nd=(/size(data,1),size(data,2), var%ihxb_end(izs) -var%ihxb_beg(izs)+1, &
     var%ihyb_end(izs) -var%ihyb_beg(izs)+1/)
     iplan=find_plan((/nd(1),nd(2),nd(3),nd(4),1/),ith)
     call fft_2pass(planI(:,iplan,ith),iws,izs,ith,data,nd)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weiift
end module wei_fft
