module Off2Ang3D_mod

  use off2ang3d_types

  use fft_data_mod

  implicit none

  logical, private :: ang_2d
  logical, private :: apply_jac_gamma,apply_jac_azim_1,apply_jac_azim_2
  logical, private :: apply_jac_both,apply_jac_none
  logical, private :: jac_nowrap
  character(len=128),private :: tag_output
  integer, external ::  sreed,srite,sseek_block

  integer, private :: n_kz,n_kmx,n_kmy,n_khx,n_khy
  integer, private :: n_hx,n_hy,n_hx_pad,n_hy_pad
  integer, private :: n_bytes_hx,n_bytes_gamma_azim
  integer, private :: n_threads,n_gamma,n_azim,number_aliases
  real, private :: sign_khy
  real, private :: hx_0,hy_0,d_hx,d_hy
  real, private :: d_kz,d_kmx,d_kmy,d_khx,d_khy
  real, private :: kz_0,kmx_0,kmy_0,khx_0,khy_0
  real, private :: azim_0,azim_0_max,gamma_0,d_azim,d_gamma
  real, private :: pi,inv_2pi,inv_pi,delta_azim_gamma_0

  integer, allocatable, dimension (:,:), private :: azim_bound
  real, allocatable, dimension (:), private :: k_mx,k_my,k_z,k_hx,k_hy
  complex, allocatable, dimension (:,:,:), private :: hx_hy,gamma_azim
  type (fft_type), private :: fft_nhx_nhy_id

contains

  function Off2Ang3D_init(off2ang3d_par)  result(status)
    type (off2ang3d_param_type) off2ang3d_par
    integer status

    integer i_k_mx,i_k_my,i_k_z,i_k_hx,i_k_hy,ierr,i_z
    real kz,kmx,kmy,khx,khy

    status=0

    pi=acos(-1.)
    inv_2pi=1./(2.*pi)
    inv_pi=1./(pi)

    ang_2d=off2ang3d_par%ang_2d
    apply_jac_gamma=off2ang3d_par%apply_jac_gamma
    apply_jac_azim_1=off2ang3d_par%apply_jac_azim_1
    apply_jac_azim_2=off2ang3d_par%apply_jac_azim_2
    apply_jac_both=apply_jac_gamma .and. (apply_jac_azim_1.or.apply_jac_azim_2)
    apply_jac_none=.not.(apply_jac_gamma.or.apply_jac_azim_1.or.apply_jac_azim_2)
    jac_nowrap=off2ang3d_par%jac_nowrap

    n_kz=off2ang3d_par%n_kz
    n_kmx=off2ang3d_par%n_kmx
    n_kmy=off2ang3d_par%n_kmy

    n_hx=off2ang3d_par%n_hx
    n_hx_pad=off2ang3d_par%n_hx_pad
    if(mod(n_hx_pad,2) /=0) then
      write(0,*)'n_hx_pad=',n_hx_pad
      call seperr ("n_hx_pad must be even")
    end if
    n_khx=n_hx_pad

    n_hy=off2ang3d_par%n_hy
    n_hy_pad=off2ang3d_par%n_hy_pad
    if((mod(n_hy_pad,2) /=0).and. (n_hy_pad /=1)) then
      write(0,*)'n_hy_pad=',n_hy_pad
      call seperr ("n_hy_pad must be even")
    end if
    n_khy=n_hy_pad

    hx_0=off2ang3d_par%hx_0
    hy_0=off2ang3d_par%hy_0
!    hy_0=0.
    d_hx=off2ang3d_par%d_hx
    d_hy=off2ang3d_par%d_hy

    d_kz=2*pi*off2ang3d_par%d_kz
    d_kmx=2*pi*off2ang3d_par%d_kmx
    d_kmy=2*pi*off2ang3d_par%d_kmy

    d_khx=2.*pi/(d_hx*FLOAT(n_hx_pad))
    d_khy=2.*pi/(d_hy*FLOAT(n_hy_pad))

    kz_0=0.
    kmx_0=0.
    kmy_0=0.

    n_azim=off2ang3d_par%n_azim_max
    azim_0=(pi*off2ang3d_par%azim_0/180.)
    d_azim=(pi*off2ang3d_par%d_azim/180.)
    azim_0_max=(pi*off2ang3d_par%azim_0_max/180.)

    n_gamma=off2ang3d_par%n_gamma
    gamma_0=(pi*off2ang3d_par%gamma_0/180.)
    d_gamma=(pi*off2ang3d_par%d_gamma/180.)


    n_threads=off2ang3d_par%n_threads
!    if(off2ang3d_par%create_coplanar_stack) then
      if(mod(n_kz,n_threads) /=0) then
        write(0,*)'n_threads=',n_threads
        write(0,*)'n_kz=',n_kz
        call seperr ("n_kz be multiple of n_threads")
      end if
!    end if

    n_bytes_hx=n_hx*n_threads*8
    if(off2ang3d_par%create_coplanar_ang) then
      n_bytes_gamma_azim=n_gamma*1*n_threads*8
    else
      n_bytes_gamma_azim=n_gamma*n_azim*n_threads*8
    end if

    allocate (azim_bound(2,n_gamma),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of azim_bound")

    allocate (k_z(n_kz),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_z ")
    allocate (k_mx(n_kmx),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
    allocate (k_my(n_kmy),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_my ")
    allocate (k_hx(n_khx),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")
    allocate (k_hy(n_khy),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hy ")


    allocate (hx_hy(n_hx_pad,n_hy_pad,n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of hx_khy ")


    allocate (gamma_azim(n_gamma,n_azim,n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of gamma_azim ")


    do i_k_z=1,n_kz
       kz  = i_k_z - 1.
       if(kz <= float(n_kz/2)) then
!          k_z(i_k_z) =  kz_0+d_kz* kz
          k_z(i_k_z) =  -(kz_0+d_kz* kz)
       else
!          k_z(i_k_z) = -kz_0+d_kz*(kz-n_kz)
          k_z(i_k_z) = -(kz_0+d_kz*(kz-n_kz))
       end if
    end do

    do i_k_mx=1,n_kmx
       kmx  = i_k_mx - 1.
       if(kmx <= float(n_kmx/2)) then
!          k_mx(i_k_mx) =  kmx_0+d_kmx* kmx
          k_mx(i_k_mx) =  -(kmx_0+d_kmx* kmx)
       else
!          k_mx(i_k_mx) = -kmx_0+d_kmx*(kmx-n_kmx)
          k_mx(i_k_mx) = -(kmx_0+d_kmx*(kmx-n_kmx))
       end if
    end do

    do i_k_my=1,n_kmy
       kmy  = i_k_my - 1.
       if(kmy <= float(n_kmy/2)) then
!          k_my(i_k_my) =  kmy_0+d_kmy* kmy
          k_my(i_k_my) =  -(kmy_0+d_kmy* kmy)
       else
!          k_my(i_k_my) = -kmy_0+d_kmy*(kmy-n_kmy)
          k_my(i_k_my) = -(kmy_0+d_kmy*(kmy-n_kmy))
       end if
    end do

    khx_0=0.
    do i_k_hx=1,n_khx
       khx  = i_k_hx - 1.
       if(khx <= float(n_khx/2)) then
          k_hx(i_k_hx) =  khx_0+d_khx* khx
       else
          k_hx(i_k_hx) = -khx_0+d_khx*(khx-n_khx)
       end if
    end do

    call reorder_fft_wavenumber(FFT_FORW,n_khx,k_hx)
    khx_0=k_hx(1)

    khy_0=0.
    do i_k_hy=1,n_khy
       khy  = i_k_hy - 1.
       if(khy <= float(n_khy/2)) then
          k_hy(i_k_hy) =  khy_0+d_khy* khy
       else
          k_hy(i_k_hy) = -khy_0+d_khy*(khy-n_khy)
       end if
    end do
    call reorder_fft_wavenumber(FFT_FORW,n_khy,k_hy)
    khy_0=k_hy(1)

    fft_nhx_nhy_id=init_fft_2d(n_hx_pad,n_hy_pad,fast_init=.false.)
    if(off2ang3d_par%debug1) write(0,*)'scale (nhx_nhy)=',fft_nhx_nhy_id%scale
    status=fft_nhx_nhy_id%ierr

     number_aliases=off2ang3d_par%number_aliases
     sign_khy=off2ang3d_par%sign_khy

     if(off2ang3d_par%debug1) then
write(0,*)'status=',status
write(0,*)'number_aliases=',number_aliases
write(0,*)'n_kz=',n_kz
write(0,*)'n_kmx=',n_kmx
write(0,*)'n_kmy=',n_kmy
write(0,*)'n_hx=',n_hx
write(0,*)'n_hy=',n_hy
write(0,*)'n_hx_pad=',n_hx_pad
write(0,*)'n_hy_pad=',n_hy_pad
write(0,*)'hx_0=',hx_0
write(0,*)'hy_0=',hy_0
write(0,*)'d_hx=',d_hx
write(0,*)'d_hy=',d_hy
write(0,*)'khx_0=',khx_0
write(0,*)'khy_0=',khy_0
write(0,*)'n_bytes_hx=',n_bytes_hx
write(0,*)'n_bytes_gamma_azim=',n_bytes_gamma_azim
write(0,*)'k_hx=',k_hx
write(0,*)'k_hy=',k_hy
write(0,*)'k_mx=',k_mx
write(0,*)'k_my=',k_my
write(0,*)'k_z=',k_z
     end if


  end function Off2Ang3D_init

  function Off2Ang3D_oper(off2ang3d_par) result(status)

  type (off2ang3d_param_type) off2ang3d_par
  integer status

  logical in_bounds
  integer i_hx,i_hy,i_kmy,i_kmx,i_kz,i_khx,i_khy,i_azim,i_gamma,i_thread
  integer i_khx_copl,i_khy_copl,i_khx_copl_p1,i_khy_copl_p1
  real inv_d_khx,kkz,kkz_den,const,khx_copl,khx_copl_int,w_khx_copl
  real dkhx_copl,dkhy_copl
  real inv_d_khy,khy_copl,khy_copl_int,w_khy_copl
  real azim,sin_azim,cos_azim,k_mx_azim,k_my_azim,khx_azim,khy_azim
  real gamma,tan_gamma,cos_delta,sin_delta,tan_alpha,jac,inv_cos_gamma_sq
  real dkmx_dazim,dkmy_dazim,dkhx_dazim,dkhy_dazim,dkhx_dgamma,dkhy_dgamma
  real tan_gamma_jac,jac_azim_norm,jac_samp
  real weight_wrap_azim_1,weight_wrap_azim_2
  real norm_jac_azim_1,norm_jac_azim_2

  inv_d_khx=1./d_khx
  inv_d_khy=1./d_khy

  if(off2ang3d_par%create_coplanar_ang .or. &
     off2ang3d_par%create_coplanar_azim_ang) then
    azim_bound=comp_azim_bound(off2ang3d_par)
  end if
  if(delta_azim_gamma_0 /= 0.) then
    weight_wrap_azim_1=1.
    weight_wrap_azim_2=inv_pi*(180./delta_azim_gamma_0)
  else
    weight_wrap_azim_1=1.
    weight_wrap_azim_2=inv_pi
  end if
  norm_jac_azim_1=(pi*delta_azim_gamma_0/180.)/(2.*sqrt(d_khx**2 + d_khy**2))
  norm_jac_azim_2=.5/(2.*sqrt(d_khx**2 + d_khy**2))

  do i_kmy=1,n_kmy
  !{
  do i_kmx=1,n_kmx
  !{

  do i_kz=1,n_kz,n_threads
  !{


    do i_thread=1,n_threads
    !{
    do i_hy=1,n_hy
      if(sreed('in',hx_hy(1,i_hy,i_thread),n_bytes_hx) /= n_bytes_hx) then
          call seperr('Problems with reading input file')
      end if
      do i_hx=n_hx+1,n_hx_pad
        hx_hy(i_hx,i_hy,i_thread)=cmplx(0.,0.)
      end do
    end do
    do i_hy=n_hy+1,n_hy_pad
      do i_hx=1,n_hx_pad
        hx_hy(i_hx,i_hy,i_thread)=cmplx(0.,0.)
      end do
    end do
    !}
    end do

    gamma_azim=0.
    do i_thread=1,n_threads
    !{

      kkz=k_z(i_kz+i_thread-1)

      call fft_2d_data_par(fft_nhx_nhy_id,FFT_FORW,n_hx_pad,n_hy_pad,i_thread,hx_hy)
      call reorder_fft_2d_data_par(FFT_FORW,n_khx,n_khy,i_thread,hx_hy)


      if(off2ang3d_par%create_coplanar_ang .or. &
         off2ang3d_par%create_coplanar_azim_ang) then
      !{

        if(off2ang3d_par%create_coplanar_ang) then
          tag_output=trim(off2ang3d_par%tag_coplanar_ang)
        else
          tag_output=trim(off2ang3d_par%tag_coplanar_azim_ang)
        end if

        if((i_kz+i_thread-1 == 1)) then
          kkz_den=k_z(2)
        else
          kkz_den=kkz
        end if


        do i_gamma=1,n_gamma
        !{

          gamma=gamma_0 + (i_gamma-1)*d_gamma
          tan_gamma=tan(gamma)
          tan_gamma_jac=tan_gamma

          if(off2ang3d_par%create_coplanar_ang) then
            gamma_azim(i_gamma,1,i_thread)= 0.
          end if

          do i_azim=azim_bound(1,i_gamma),azim_bound(2,i_gamma)
          !{
            azim=azim_0_max + (i_azim-1)*d_azim
            sin_azim=sin(azim)
            cos_azim=cos(azim)
      
      
            k_mx_azim=   k_mx(i_kmx)*cos_azim + k_my(i_kmy)*sin_azim
            k_my_azim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim

! notice differrent than my notes because change in sign in beta
            dkmx_dazim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim
            dkmy_dazim= - k_mx(i_kmx)*cos_azim - k_my(i_kmy)*sin_azim

! inverse is
!      k_mx= k_mx_azim*cos_azim + k_my_azim*sin_azim
!      k_my=-k_mx_azim*sin_azim + k_my_azim*cos_azim


            const=k_my_azim*k_mx_azim/(kkz_den**2 + k_my_azim**2)

            if(ang_2d) then
              khx_azim= -tan_gamma*sign(sqrt(kkz**2),kkz)
            else
              khx_azim= -tan_gamma*sign(sqrt(kkz**2 + k_my_azim**2),kkz)
            end if

            khy_azim=-sign_khy*khx_azim*const
    
            khx_copl= khx_azim*cos_azim - khy_azim*sin_azim
            khy_copl= khx_azim*sin_azim + khy_azim*cos_azim

            if(apply_jac_none) then
            !{
              jac=1
            !}
            else
            !{

              sin_delta=k_my_azim/sqrt(k_my_azim**2 + kkz_den**2) 
              cos_delta=abs(kkz_den)/sqrt(k_my_azim**2 + kkz_den**2) 
              tan_alpha=k_mx_azim/kkz_den


              if(apply_jac_both) then

call seperr("This is not working properly yet")
                inv_cos_gamma_sq=(1.+tan_gamma_jac**2)
    
                dkhx_dazim=-tan_gamma_jac*sin_delta*dkmy_dazim
                dkhy_dazim= tan_gamma_jac*(           sin_delta*dkmx_dazim+ &
                                      tan_alpha*(cos_delta**3)*dkmy_dazim)
    
                dkhx_dgamma=kkz_den*inv_cos_gamma_sq/cos_delta
                dkhy_dgamma=kkz_den*inv_cos_gamma_sq*tan_alpha*sin_delta
    
                jac_samp=(dkhx_dgamma*dkhy_dazim + dkhy_dgamma*dkhx_dazim)*&
                          (d_azim/(d_khx*d_khy))
                jac=jac_samp


              else
                if(apply_jac_azim_1 .or. apply_jac_azim_2) then
  
                  dkhx_dazim=-tan_gamma_jac*sin_delta*dkmy_dazim
                  dkhy_dazim= tan_gamma_jac*(         sin_delta*dkmx_dazim + &
                                      tan_alpha*(cos_delta**3)*dkmy_dazim)

                  dkhx_copl = dkhx_dazim*cos_azim - dkhy_dazim*sin_azim + &
                              (- khx_azim*sin_azim - khy_azim*cos_azim)
                  dkhy_copl = dkhx_dazim*sin_azim + dkhy_dazim*cos_azim + &
                              (  khx_azim*cos_azim - khy_azim*sin_azim)

                  if(apply_jac_azim_1) then
                    if(jac_nowrap) then
                      if(gamma /= 0.) then
                        jac=max(weight_wrap_azim_1, &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_1)
                      else
                        jac=0.
                      end if
                    else
                      if(gamma /= 0.) then
                        jac=max(weight_wrap_azim_1, &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_1)
                      else
                        jac=1.
                      end if
                    end if
                  else
                    if(jac_nowrap) then
                      if(gamma /= 0.) then
                        jac=max(0., &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_2)
    
                      else
                        jac=0.
                      end if
                    else
                      if(gamma /= 0.) then
                        jac=max(weight_wrap_azim_2, &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_2)
                      else
                        jac=weight_wrap_azim_2
                      end if
                    end if
                  end if

                else
                  if(apply_jac_gamma) then
call seperr("This is not working properly yet")
                    inv_cos_gamma_sq=(1.+tan_gamma_jac**2)
    
                    dkhx_dgamma=kkz_den*inv_cos_gamma_sq/cos_delta
                    dkhy_dgamma=kkz_den*inv_cos_gamma_sq*tan_alpha*sin_delta
        
                    jac=sqrt(dkhx_dgamma**2 + dkhy_dgamma**2)
                  end if
                end if
              end if
            !}
            end if


            khx_copl_int=(khx_copl-khx_0)*inv_d_khx+1
            i_khx_copl=floor(khx_copl_int)
            w_khx_copl=khx_copl_int-i_khx_copl
  
            khy_copl_int=(khy_copl-khy_0)*inv_d_khy+1
            i_khy_copl=floor(khy_copl_int)
            w_khy_copl=khy_copl_int-i_khy_copl
  
            if((i_khx_copl > -n_khx*number_aliases) .and. &
               (i_khx_copl <= (n_khx*(number_aliases+1)) ) .and.  &
               (i_khy_copl > -n_khy*number_aliases) .and. &
               (i_khy_copl <= (n_khy*(number_aliases+1) )) &
              ) then
              in_bounds=.true.
              i_khx_copl=   imod((number_aliases+1)*n_khx+i_khx_copl-1,n_khx)+1
              i_khx_copl_p1=imod((number_aliases+1)*n_khx+i_khx_copl  ,n_khx)+1
  
              i_khy_copl=   imod((number_aliases+1)*n_khy+i_khy_copl-1,n_khy)+1
              i_khy_copl_p1=imod((number_aliases+1)*n_khy+i_khy_copl  ,n_khy)+1


            else
                in_bounds=.false.
            end if

            if(in_bounds) then
            !{

if(off2ang3d_par%create_coplanar_ang) then

  gamma_azim(i_gamma,1,i_thread)=  gamma_azim(i_gamma,1,i_thread) +  &
  (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
  ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0))* jac

else
  gamma_azim(i_gamma,i_azim,i_thread)=   &
  (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
  ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0)) * jac

end if

            !}
            end if
          !} azim
          end do
        !} gamma
        end do
      !}
      else
      !{
call seperr("This is not working properly yet")
        do i_azim=1,n_azim
        !{

          azim=azim_0 + (i_azim-1)*d_azim
          sin_azim=sin(azim)
          cos_azim=cos(azim)


          k_mx_azim=   k_mx(i_kmx)*cos_azim + k_my(i_kmy)*sin_azim
          k_my_azim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim
  
! notice differrent than my notes because change in sign in beta
          dkmx_dazim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim
          dkmy_dazim= - k_mx(i_kmx)*cos_azim - k_my(i_kmy)*sin_azim

! inverse is
!      k_mx= k_mx_azim*cos_azim + k_my_azim*sin_azim
!      k_my=-k_mx_azim*sin_azim + k_my_azim*cos_azim


        tag_output=trim(off2ang3d_par%tag_coplanar_stack)
        if(off2ang3d_par%create_stack) then
        !{
          gamma_azim(1,i_azim,i_thread)=cmplx(0.,0.)
          do i_khy=1,n_khy
          !{
  
            do i_khx=1,n_khx
            !{
              gamma_azim(1,i_azim,i_thread)= gamma_azim(1,i_azim,i_thread) + &
                  hx_hy(i_khx,i_khy,i_thread)* &
                  cexp(cmplx(0,k_hx(i_khx)*hx_0+k_hy(i_khy)*hy_0))* &
                  fft_nhx_nhy_id%scale
  
            !}
            end do
  
          !}
          end do
  
        !}
        else
        !{
          gamma_azim(1,i_azim,i_thread)=cmplx(0.,0.)
  
  !        do i_khy=1,n_khy
  !        !{
  
            if((i_kz+i_thread-1 == 1)) then
              kkz_den=k_z(2)
            else
              kkz_den=kkz
            end if

            if(.not. apply_jac_none) then
              sin_delta=k_my_azim/sqrt(k_my_azim**2 + kkz_den**2) 
              cos_delta=abs(kkz_den)/sqrt(k_my_azim**2 + kkz_den**2) 
              tan_alpha=k_mx_azim/kkz_den
            end if

            const=-k_my_azim*k_mx_azim/(kkz_den**2 + k_my_azim**2)
  
            do i_khx=1,n_khx
            !{

              khx_azim=k_hx(i_khx)
              khy_azim=sign_khy*khx_azim*const

              if(apply_jac_none) then
                jac=1
              else
                tan_gamma_jac=(khx_azim*cos_delta)/kkz_den
                if(apply_jac_azim_1 .or. apply_jac_azim_2) then
    
                  dkhx_dazim=tan_gamma_jac*sin_delta*dkmy_dazim
                  dkhy_dazim=tan_gamma_jac*(           sin_delta*dkmx_dazim -&
                                        tan_alpha*(cos_delta**3)*dkmy_dazim)
      
                  jac_samp=sqrt((dkhx_dazim*d_azim/d_khx)**2 + &
                                (dkhy_dazim*d_azim/d_khy)**2 )
                  jac=jac_samp
                else
                  if(apply_jac_gamma) then
                    call seperr('Not possible to apply_jac_gamma in this case')
                  else
                    if(apply_jac_both) then
                    call seperr('Not possible to apply_jac_both in this case')
                    end if
                  end if
                end if
              end if

  

              khx_copl= khx_azim*cos_azim - khy_azim*sin_azim
              khy_copl= khx_azim*sin_azim + khy_azim*cos_azim
  
              khx_copl_int=(khx_copl-khx_0)*inv_d_khx+1
              i_khx_copl=floor(khx_copl_int)
              w_khx_copl=khx_copl_int-i_khx_copl
  
              khy_copl_int=(khy_copl-khy_0)*inv_d_khy+1
              i_khy_copl=floor(khy_copl_int)
              w_khy_copl=khy_copl_int-i_khy_copl
  
              if((i_khx_copl > -n_khx*number_aliases) .and. &
                 (i_khx_copl <= (n_khx*(number_aliases+1)) ) .and.  &
                 (i_khy_copl > -n_khy*number_aliases) .and. &
                 (i_khy_copl <= (n_khy*(number_aliases+1) )) &
                ) then
                in_bounds=.true.
            i_khx_copl=   imod((number_aliases+1)*n_khx+i_khx_copl-1,n_khx)+1
            i_khx_copl_p1=imod((number_aliases+1)*n_khx+i_khx_copl  ,n_khx)+1

            i_khy_copl=   imod((number_aliases+1)*n_khy+i_khy_copl-1,n_khy)+1
            i_khy_copl_p1=imod((number_aliases+1)*n_khy+i_khy_copl  ,n_khy)+1
              else
                in_bounds=.false.
              end if
  
              if(in_bounds) then
              !{

                gamma_azim(1,i_azim,i_thread)= gamma_azim(1,i_azim,i_thread)+&
 (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+&
  hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+&
  hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+&
  hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
     ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0)) * jac
              !}
              end if
  
            !} khx
            end do
  
  !        !} khy
  !        end do
          !}
          end if

        !} azim
        end do

        !}
        end if

      !} thread
      end do

    if(srite(trim(tag_output), &
         gamma_azim,n_bytes_gamma_azim)/= n_bytes_gamma_azim) then
        write(0,*)'tag_output=',trim(tag_output)
        call seperr('Problems with writing tag_output file')
    end if

  !} kz
  end do

  !} kmx
  end do
  !} kmy
  end do
        

  status=0
      
  end function Off2Ang3D_oper


subroutine fft_2d_data_par(fft_id,sign,n1,n2,i_thread,data_par)
type (fft_type) fft_id
integer :: n1,n2,i_thread
integer :: sign
complex, dimension (:,:,:) :: data_par

call fft_2d_data(fft_id,sign,n1,n2,data_par(:,:,i_thread))
!
end subroutine fft_2d_data_par

subroutine reorder_fft_2d_data_par(sign,n1,n2,i_thread,data_par)
integer :: n1,n2,i_thread
integer :: sign
complex, dimension (:,:,:) :: data_par

integer n1d2,i1,n2d2,i2
complex, dimension (n1) :: temp1
complex, dimension (n2) :: temp2

if(n1 /= 1) then
  n1d2=n1/2
else
  n1d2=1
end if

if(n2 /= 1) then
  n2d2=n2/2
else
  n2d2=1
end if

if(sign==FFT_FORW) then

do i2=1,n2
  do i1=n1d2,n1
    temp1(i1)=data_par(i1-n1d2+1,i2,i_thread)
  end do
  do i1=1,n1d2-1
    temp1(i1)=data_par(n1d2+i1+1,i2,i_thread)
  end do

  do i1=1,n1
    data_par(i1,i2,i_thread)=temp1(i1)
  end do
end do

do i1=1,n1
  do i2=n2d2,n2
    temp2(i2)=data_par(i1,i2-n2d2+1,i_thread)
  end do
  do i2=1,n2d2-1
    temp2(i2)=data_par(i1,n2d2+i2+1,i_thread)
  end do

  do i2=1,n2
    data_par(i1,i2,i_thread)=temp2(i2)
  end do
end do

end if


end subroutine reorder_fft_2d_data_par

subroutine reorder_fft_wavenumber(sign,n1,k)
integer :: n1
integer :: sign
real, dimension (:) :: k

integer n1d2,i1
real, dimension (n1) :: temp

if(n1 /= 1) then
  n1d2=n1/2
else
  n1d2=1
end if

if(sign==FFT_FORW) then

  do i1=n1d2,n1
    temp(i1)=k(i1-n1d2+1)
  end do
  do i1=1,n1d2-1
    temp(i1)=k(n1d2+i1+1)
  end do

  do i1=1,n1
    k(i1)=temp(i1)
  end do


end if

end subroutine reorder_fft_wavenumber

function  comp_azim_bound(off2ang3d_par) result (azim_bound)
  type (off2ang3d_param_type) off2ang3d_par
  integer azim_bound(2,off2ang3d_par%n_gamma)


  integer power,i_gamma
  real azim_min,azim_max,azim_min_0,azim_max_0
  real inv_d_azim,gamma_max_max,conv
  real azim_bound_min,azim_bound_max,delta_gamma,gamma


  pi=acos(-1.)
  if(off2ang3d_par%azim_min_gamma_max /= 0.) then
    azim_min=off2ang3d_par%azim_min_gamma_max
  else
    azim_min=off2ang3d_par%azim_0
  end if
  if(off2ang3d_par%azim_max_gamma_max /= 0.) then
    azim_max=off2ang3d_par%azim_max_gamma_max
  else
    azim_max=off2ang3d_par%azim_0+(off2ang3d_par%n_azim-1)*off2ang3d_par%d_azim
  end if

  if(off2ang3d_par%azim_gamma_0 /= 0.) then
    azim_min_0=-off2ang3d_par%azim_gamma_0
    azim_max_0=off2ang3d_par%azim_gamma_0
  else
    azim_min_0=azim_min
    azim_max_0=azim_max
  end if
  inv_d_azim=1./off2ang3d_par%d_azim
  power=off2ang3d_par%sin_pow
  delta_azim_gamma_0=azim_max_0-azim_min_0

if(off2ang3d_par%debug1) then
  write(0,*)'azim_min=',azim_min
  write(0,*)'azim_max=',azim_max
  write(0,*)'delta_azim_gamma_0=',delta_azim_gamma_0
  write(0,*)'inv_d_azim=',inv_d_azim
end if

  gamma_max_max=90
  delta_gamma=gamma_max_max
  conv=pi/2
  do i_gamma=1,off2ang3d_par%n_gamma
    gamma=off2ang3d_par%gamma_0+(i_gamma-1)*off2ang3d_par%d_gamma

!    if(gamma >0.) then
      azim_bound_min=azim_min + &
        (azim_min_0-azim_min)*(sin(conv*(gamma_max_max-gamma)/delta_gamma)**power)
      azim_bound_max=azim_max + &
        (azim_max_0-azim_max)*(sin(conv*(gamma_max_max-gamma)/delta_gamma)**power)
      azim_bound(1,i_gamma)=(azim_bound_min-off2ang3d_par%azim_0_max)*inv_d_azim+1
      azim_bound(2,i_gamma)=(azim_bound_max-off2ang3d_par%azim_0_max)*inv_d_azim+1

if(off2ang3d_par%debug1) then
  write(0,*)'gamma=',gamma
  write(0,*)'azim_0_max=',off2ang3d_par%azim_0_max
  write(0,*)'azim_bound_min=',azim_bound_min
  write(0,*)'azim_bound_max=',azim_bound_max
  write(0,*)'azim_bound(1)=',azim_bound(1,i_gamma)
  write(0,*)'azim_bound(2)=',azim_bound(2,i_gamma)
end if

   end do


end function comp_azim_bound


end module Off2Ang3D_mod
